# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2022-2023 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Component registry helpers."""

import os
import shutil
import tempfile
from distutils.dir_util import copy_tree  # pylint: disable=deprecated-module
from pathlib import Path
from shutil import copytree
from typing import Optional, cast

import click
from aea.cli.registry.settings import REGISTRY_LOCAL, REGISTRY_REMOTE, REMOTE_IPFS
from aea.cli.utils.click_utils import reraise_as_click_exception
from aea.cli.utils.config import (
    get_default_remote_registry,
    get_ipfs_node_multiaddr,
    load_item_config,
)
from aea.cli.utils.context import Context
from aea.cli.utils.package_utils import (
    try_get_item_source_path,
    try_get_item_target_path,
)
from aea.configurations.constants import DEFAULT_README_FILE, SERVICE, SERVICES
from aea.configurations.data_types import PublicId
from aea.helpers.cid import to_v1

from autonomy.configurations.base import (
    DEFAULT_SERVICE_CONFIG_FILE,
    PACKAGE_TYPE_TO_CONFIG_CLASS,
)


try:
    from aea_cli_ipfs.ipfs_utils import IPFSTool  # type: ignore

    IS_IPFS_PLUGIN_INSTALLED = True
except ImportError:  # pragma: nocover
    IS_IPFS_PLUGIN_INSTALLED = False


def fetch_service(
    ctx: Context,
    public_id: PublicId,
    alias: Optional[str] = None,
) -> Path:
    """Fetch service."""

    if ctx.registry_type == REGISTRY_REMOTE:
        return fetch_service_remote(public_id, alias=alias)
    if ctx.registry_type == REGISTRY_LOCAL:
        return fetch_service_local(ctx, public_id, alias=alias)
    return fetch_service_mixed(ctx, public_id, alias=alias)


def fetch_service_mixed(
    ctx: Context,
    public_id: PublicId,
    alias: Optional[str] = None,
) -> Path:
    """Fetch service in mixed mode."""
    try:
        return fetch_service_local(ctx, public_id, alias=alias)
    except Exception as e:  # pylint: disable=broad-except
        click.echo(
            f"Fetch from local registry failed (reason={str(e)}), trying remote registry..."
        )
        return fetch_service_remote(public_id, alias=alias)


def fetch_service_remote(
    public_id: PublicId,
    alias: Optional[str] = None,
) -> Path:
    """Fetch service in remote mode."""
    if get_default_remote_registry() == REMOTE_IPFS:
        return fetch_service_ipfs(public_id, alias=alias)

    raise Exception("HTTP registry not supported.")  # pragma: nocover


def fetch_service_ipfs(
    public_id: PublicId,
    alias: Optional[str] = None,
) -> Path:
    """Fetch service from IPFS node."""

    if not IS_IPFS_PLUGIN_INSTALLED:
        raise RuntimeError("IPFS plugin not installed.")  # pragma: no cover

    with tempfile.TemporaryDirectory() as temp_dir:
        ipfs_tool = IPFSTool(get_ipfs_node_multiaddr())
        download_path = Path(ipfs_tool.download(public_id.hash, temp_dir))
        package_path = Path.cwd() / (alias or download_path.name)
        shutil.copytree(download_path, package_path)

    if not Path(package_path, DEFAULT_SERVICE_CONFIG_FILE).exists():
        raise click.ClickException(
            "Downloaded packages is not a service package, "
            "if you intend to download an agent please use `--agent` flag "
            "or check the hash"
        )

    service_config = load_item_config(
        SERVICE, package_path, PACKAGE_TYPE_TO_CONFIG_CLASS
    )

    click.echo(
        f"Downloaded service package {service_config.public_id} @ {package_path}"
    )

    return package_path


def fetch_service_local(
    ctx: Context,
    public_id: PublicId,
    alias: Optional[str] = None,
) -> Path:
    """Fetch service from local directory."""

    with reraise_as_click_exception(ValueError):
        registry_path = ctx.registry_path

    source_path = try_get_item_source_path(
        registry_path, public_id.author, SERVICES, public_id.name
    )

    target_path = Path(ctx.cwd, alias or public_id.name)
    if target_path.exists():
        raise click.ClickException(
            f'Item "{target_path.name}" already exists in target folder "{target_path.parent}".'
        )

    copy_tree(source_path, str(target_path))
    click.echo(f"Copied service package {public_id}")
    return target_path


def publish_service_package(click_context: click.Context, registry: str) -> None:
    """Publish a service package."""

    # TODO ensure we have error handling here.
    service_config = load_item_config(
        SERVICE, Path(click_context.obj.cwd), PACKAGE_TYPE_TO_CONFIG_CLASS
    )

    if registry == REGISTRY_REMOTE:
        if get_default_remote_registry() == REMOTE_IPFS:
            publish_service_ipfs(service_config.public_id, Path(click_context.obj.cwd))
        else:
            raise Exception("HTTP registry not supported.")  # pragma: no cover

    else:
        publish_service_local(
            cast(
                Context,
                click_context.obj,
            ),
            service_config.public_id,
        )


def publish_service_ipfs(public_id: PublicId, package_path: Path) -> None:
    """Publish a service package on the IPFS registry."""

    if not IS_IPFS_PLUGIN_INSTALLED:  # pragma: nocover
        raise RuntimeError("IPFS plugin not installed.")

    package_path = package_path.resolve()

    with tempfile.TemporaryDirectory() as temp_dir:
        temp_service_dir = Path(temp_dir, public_id.name)
        temp_service_dir.mkdir()
        shutil.copyfile(
            package_path / DEFAULT_SERVICE_CONFIG_FILE,
            temp_service_dir / DEFAULT_SERVICE_CONFIG_FILE,
        )
        shutil.copyfile(
            package_path / DEFAULT_README_FILE,
            temp_service_dir / DEFAULT_README_FILE,
        )

        ipfs_tool = IPFSTool(get_ipfs_node_multiaddr())
        _, package_hash, _ = ipfs_tool.add(str(temp_service_dir.resolve()))
        package_hash = to_v1(package_hash)

        click.echo(
            f'Service "{public_id.name}" successfully published on the IPFS registry.\n\tPublicId: {public_id}\n\tPackage hash: {package_hash}'
        )


def publish_service_local(ctx: Context, public_id: PublicId) -> None:
    """Publish a service package on the local packages directory."""

    with reraise_as_click_exception(ValueError):
        registry_path = ctx.registry_path

    target_dir = try_get_item_target_path(
        registry_path,
        public_id.author,
        SERVICES,
        public_id.name,
    )
    author_dir = Path(target_dir).parent
    if not os.path.exists(author_dir):
        os.makedirs(author_dir, exist_ok=True)
    # TODO: also make services dir?

    copytree(ctx.cwd, target_dir)
    click.echo(
        f'Service "{public_id.name}" successfully published on the local packages directory.'
    )



# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""
import grpc

from packages.valory.connections.abci.tendermint.abci import (
    types_pb2 as tendermint_dot_abci_dot_types__pb2,
)


class ABCIApplicationStub(object):
    """----------------------------------------
    Service Definition

    """

    def __init__(self, channel):
        """Constructor.

        :param channel: A grpc.Channel.
        """
        self.Echo = channel.unary_unary(
            "/tendermint.abci.ABCIApplication/Echo",
            request_serializer=tendermint_dot_abci_dot_types__pb2.RequestEcho.SerializeToString,
            response_deserializer=tendermint_dot_abci_dot_types__pb2.ResponseEcho.FromString,
        )
        self.Flush = channel.unary_unary(
            "/tendermint.abci.ABCIApplication/Flush",
            request_serializer=tendermint_dot_abci_dot_types__pb2.RequestFlush.SerializeToString,
            response_deserializer=tendermint_dot_abci_dot_types__pb2.ResponseFlush.FromString,
        )
        self.Info = channel.unary_unary(
            "/tendermint.abci.ABCIApplication/Info",
            request_serializer=tendermint_dot_abci_dot_types__pb2.RequestInfo.SerializeToString,
            response_deserializer=tendermint_dot_abci_dot_types__pb2.ResponseInfo.FromString,
        )
        self.SetOption = channel.unary_unary(
            "/tendermint.abci.ABCIApplication/SetOption",
            request_serializer=tendermint_dot_abci_dot_types__pb2.RequestSetOption.SerializeToString,
            response_deserializer=tendermint_dot_abci_dot_types__pb2.ResponseSetOption.FromString,
        )
        self.DeliverTx = channel.unary_unary(
            "/tendermint.abci.ABCIApplication/DeliverTx",
            request_serializer=tendermint_dot_abci_dot_types__pb2.RequestDeliverTx.SerializeToString,
            response_deserializer=tendermint_dot_abci_dot_types__pb2.ResponseDeliverTx.FromString,
        )
        self.CheckTx = channel.unary_unary(
            "/tendermint.abci.ABCIApplication/CheckTx",
            request_serializer=tendermint_dot_abci_dot_types__pb2.RequestCheckTx.SerializeToString,
            response_deserializer=tendermint_dot_abci_dot_types__pb2.ResponseCheckTx.FromString,
        )
        self.Query = channel.unary_unary(
            "/tendermint.abci.ABCIApplication/Query",
            request_serializer=tendermint_dot_abci_dot_types__pb2.RequestQuery.SerializeToString,
            response_deserializer=tendermint_dot_abci_dot_types__pb2.ResponseQuery.FromString,
        )
        self.Commit = channel.unary_unary(
            "/tendermint.abci.ABCIApplication/Commit",
            request_serializer=tendermint_dot_abci_dot_types__pb2.RequestCommit.SerializeToString,
            response_deserializer=tendermint_dot_abci_dot_types__pb2.ResponseCommit.FromString,
        )
        self.InitChain = channel.unary_unary(
            "/tendermint.abci.ABCIApplication/InitChain",
            request_serializer=tendermint_dot_abci_dot_types__pb2.RequestInitChain.SerializeToString,
            response_deserializer=tendermint_dot_abci_dot_types__pb2.ResponseInitChain.FromString,
        )
        self.BeginBlock = channel.unary_unary(
            "/tendermint.abci.ABCIApplication/BeginBlock",
            request_serializer=tendermint_dot_abci_dot_types__pb2.RequestBeginBlock.SerializeToString,
            response_deserializer=tendermint_dot_abci_dot_types__pb2.ResponseBeginBlock.FromString,
        )
        self.EndBlock = channel.unary_unary(
            "/tendermint.abci.ABCIApplication/EndBlock",
            request_serializer=tendermint_dot_abci_dot_types__pb2.RequestEndBlock.SerializeToString,
            response_deserializer=tendermint_dot_abci_dot_types__pb2.ResponseEndBlock.FromString,
        )
        self.ListSnapshots = channel.unary_unary(
            "/tendermint.abci.ABCIApplication/ListSnapshots",
            request_serializer=tendermint_dot_abci_dot_types__pb2.RequestListSnapshots.SerializeToString,
            response_deserializer=tendermint_dot_abci_dot_types__pb2.ResponseListSnapshots.FromString,
        )
        self.OfferSnapshot = channel.unary_unary(
            "/tendermint.abci.ABCIApplication/OfferSnapshot",
            request_serializer=tendermint_dot_abci_dot_types__pb2.RequestOfferSnapshot.SerializeToString,
            response_deserializer=tendermint_dot_abci_dot_types__pb2.ResponseOfferSnapshot.FromString,
        )
        self.LoadSnapshotChunk = channel.unary_unary(
            "/tendermint.abci.ABCIApplication/LoadSnapshotChunk",
            request_serializer=tendermint_dot_abci_dot_types__pb2.RequestLoadSnapshotChunk.SerializeToString,
            response_deserializer=tendermint_dot_abci_dot_types__pb2.ResponseLoadSnapshotChunk.FromString,
        )
        self.ApplySnapshotChunk = channel.unary_unary(
            "/tendermint.abci.ABCIApplication/ApplySnapshotChunk",
            request_serializer=tendermint_dot_abci_dot_types__pb2.RequestApplySnapshotChunk.SerializeToString,
            response_deserializer=tendermint_dot_abci_dot_types__pb2.ResponseApplySnapshotChunk.FromString,
        )


class ABCIApplicationServicer(object):
    """----------------------------------------
    Service Definition

    """

    def Echo(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def Flush(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def Info(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def SetOption(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def DeliverTx(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def CheckTx(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def Query(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def Commit(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def InitChain(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def BeginBlock(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def EndBlock(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def ListSnapshots(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def OfferSnapshot(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def LoadSnapshotChunk(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def ApplySnapshotChunk(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")


def add_ABCIApplicationServicer_to_server(servicer, server):
    rpc_method_handlers = {
        "Echo": grpc.unary_unary_rpc_method_handler(
            servicer.Echo,
            request_deserializer=tendermint_dot_abci_dot_types__pb2.RequestEcho.FromString,
            response_serializer=tendermint_dot_abci_dot_types__pb2.ResponseEcho.SerializeToString,
        ),
        "Flush": grpc.unary_unary_rpc_method_handler(
            servicer.Flush,
            request_deserializer=tendermint_dot_abci_dot_types__pb2.RequestFlush.FromString,
            response_serializer=tendermint_dot_abci_dot_types__pb2.ResponseFlush.SerializeToString,
        ),
        "Info": grpc.unary_unary_rpc_method_handler(
            servicer.Info,
            request_deserializer=tendermint_dot_abci_dot_types__pb2.RequestInfo.FromString,
            response_serializer=tendermint_dot_abci_dot_types__pb2.ResponseInfo.SerializeToString,
        ),
        "SetOption": grpc.unary_unary_rpc_method_handler(
            servicer.SetOption,
            request_deserializer=tendermint_dot_abci_dot_types__pb2.RequestSetOption.FromString,
            response_serializer=tendermint_dot_abci_dot_types__pb2.ResponseSetOption.SerializeToString,
        ),
        "DeliverTx": grpc.unary_unary_rpc_method_handler(
            servicer.DeliverTx,
            request_deserializer=tendermint_dot_abci_dot_types__pb2.RequestDeliverTx.FromString,
            response_serializer=tendermint_dot_abci_dot_types__pb2.ResponseDeliverTx.SerializeToString,
        ),
        "CheckTx": grpc.unary_unary_rpc_method_handler(
            servicer.CheckTx,
            request_deserializer=tendermint_dot_abci_dot_types__pb2.RequestCheckTx.FromString,
            response_serializer=tendermint_dot_abci_dot_types__pb2.ResponseCheckTx.SerializeToString,
        ),
        "Query": grpc.unary_unary_rpc_method_handler(
            servicer.Query,
            request_deserializer=tendermint_dot_abci_dot_types__pb2.RequestQuery.FromString,
            response_serializer=tendermint_dot_abci_dot_types__pb2.ResponseQuery.SerializeToString,
        ),
        "Commit": grpc.unary_unary_rpc_method_handler(
            servicer.Commit,
            request_deserializer=tendermint_dot_abci_dot_types__pb2.RequestCommit.FromString,
            response_serializer=tendermint_dot_abci_dot_types__pb2.ResponseCommit.SerializeToString,
        ),
        "InitChain": grpc.unary_unary_rpc_method_handler(
            servicer.InitChain,
            request_deserializer=tendermint_dot_abci_dot_types__pb2.RequestInitChain.FromString,
            response_serializer=tendermint_dot_abci_dot_types__pb2.ResponseInitChain.SerializeToString,
        ),
        "BeginBlock": grpc.unary_unary_rpc_method_handler(
            servicer.BeginBlock,
            request_deserializer=tendermint_dot_abci_dot_types__pb2.RequestBeginBlock.FromString,
            response_serializer=tendermint_dot_abci_dot_types__pb2.ResponseBeginBlock.SerializeToString,
        ),
        "EndBlock": grpc.unary_unary_rpc_method_handler(
            servicer.EndBlock,
            request_deserializer=tendermint_dot_abci_dot_types__pb2.RequestEndBlock.FromString,
            response_serializer=tendermint_dot_abci_dot_types__pb2.ResponseEndBlock.SerializeToString,
        ),
        "ListSnapshots": grpc.unary_unary_rpc_method_handler(
            servicer.ListSnapshots,
            request_deserializer=tendermint_dot_abci_dot_types__pb2.RequestListSnapshots.FromString,
            response_serializer=tendermint_dot_abci_dot_types__pb2.ResponseListSnapshots.SerializeToString,
        ),
        "OfferSnapshot": grpc.unary_unary_rpc_method_handler(
            servicer.OfferSnapshot,
            request_deserializer=tendermint_dot_abci_dot_types__pb2.RequestOfferSnapshot.FromString,
            response_serializer=tendermint_dot_abci_dot_types__pb2.ResponseOfferSnapshot.SerializeToString,
        ),
        "LoadSnapshotChunk": grpc.unary_unary_rpc_method_handler(
            servicer.LoadSnapshotChunk,
            request_deserializer=tendermint_dot_abci_dot_types__pb2.RequestLoadSnapshotChunk.FromString,
            response_serializer=tendermint_dot_abci_dot_types__pb2.ResponseLoadSnapshotChunk.SerializeToString,
        ),
        "ApplySnapshotChunk": grpc.unary_unary_rpc_method_handler(
            servicer.ApplySnapshotChunk,
            request_deserializer=tendermint_dot_abci_dot_types__pb2.RequestApplySnapshotChunk.FromString,
            response_serializer=tendermint_dot_abci_dot_types__pb2.ResponseApplySnapshotChunk.SerializeToString,
        ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
        "tendermint.abci.ABCIApplication", rpc_method_handlers
    )
    server.add_generic_rpc_handlers((generic_handler,))


# This class is part of an EXPERIMENTAL API.
class ABCIApplication(object):
    """----------------------------------------
    Service Definition

    """

    @staticmethod
    def Echo(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/tendermint.abci.ABCIApplication/Echo",
            tendermint_dot_abci_dot_types__pb2.RequestEcho.SerializeToString,
            tendermint_dot_abci_dot_types__pb2.ResponseEcho.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )

    @staticmethod
    def Flush(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/tendermint.abci.ABCIApplication/Flush",
            tendermint_dot_abci_dot_types__pb2.RequestFlush.SerializeToString,
            tendermint_dot_abci_dot_types__pb2.ResponseFlush.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )

    @staticmethod
    def Info(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/tendermint.abci.ABCIApplication/Info",
            tendermint_dot_abci_dot_types__pb2.RequestInfo.SerializeToString,
            tendermint_dot_abci_dot_types__pb2.ResponseInfo.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )

    @staticmethod
    def SetOption(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/tendermint.abci.ABCIApplication/SetOption",
            tendermint_dot_abci_dot_types__pb2.RequestSetOption.SerializeToString,
            tendermint_dot_abci_dot_types__pb2.ResponseSetOption.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )

    @staticmethod
    def DeliverTx(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/tendermint.abci.ABCIApplication/DeliverTx",
            tendermint_dot_abci_dot_types__pb2.RequestDeliverTx.SerializeToString,
            tendermint_dot_abci_dot_types__pb2.ResponseDeliverTx.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )

    @staticmethod
    def CheckTx(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/tendermint.abci.ABCIApplication/CheckTx",
            tendermint_dot_abci_dot_types__pb2.RequestCheckTx.SerializeToString,
            tendermint_dot_abci_dot_types__pb2.ResponseCheckTx.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )

    @staticmethod
    def Query(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/tendermint.abci.ABCIApplication/Query",
            tendermint_dot_abci_dot_types__pb2.RequestQuery.SerializeToString,
            tendermint_dot_abci_dot_types__pb2.ResponseQuery.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )

    @staticmethod
    def Commit(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/tendermint.abci.ABCIApplication/Commit",
            tendermint_dot_abci_dot_types__pb2.RequestCommit.SerializeToString,
            tendermint_dot_abci_dot_types__pb2.ResponseCommit.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )

    @staticmethod
    def InitChain(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/tendermint.abci.ABCIApplication/InitChain",
            tendermint_dot_abci_dot_types__pb2.RequestInitChain.SerializeToString,
            tendermint_dot_abci_dot_types__pb2.ResponseInitChain.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )

    @staticmethod
    def BeginBlock(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/tendermint.abci.ABCIApplication/BeginBlock",
            tendermint_dot_abci_dot_types__pb2.RequestBeginBlock.SerializeToString,
            tendermint_dot_abci_dot_types__pb2.ResponseBeginBlock.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )

    @staticmethod
    def EndBlock(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/tendermint.abci.ABCIApplication/EndBlock",
            tendermint_dot_abci_dot_types__pb2.RequestEndBlock.SerializeToString,
            tendermint_dot_abci_dot_types__pb2.ResponseEndBlock.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )

    @staticmethod
    def ListSnapshots(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/tendermint.abci.ABCIApplication/ListSnapshots",
            tendermint_dot_abci_dot_types__pb2.RequestListSnapshots.SerializeToString,
            tendermint_dot_abci_dot_types__pb2.ResponseListSnapshots.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )

    @staticmethod
    def OfferSnapshot(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/tendermint.abci.ABCIApplication/OfferSnapshot",
            tendermint_dot_abci_dot_types__pb2.RequestOfferSnapshot.SerializeToString,
            tendermint_dot_abci_dot_types__pb2.ResponseOfferSnapshot.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )

    @staticmethod
    def LoadSnapshotChunk(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/tendermint.abci.ABCIApplication/LoadSnapshotChunk",
            tendermint_dot_abci_dot_types__pb2.RequestLoadSnapshotChunk.SerializeToString,
            tendermint_dot_abci_dot_types__pb2.ResponseLoadSnapshotChunk.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )

    @staticmethod
    def ApplySnapshotChunk(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/tendermint.abci.ABCIApplication/ApplySnapshotChunk",
            tendermint_dot_abci_dot_types__pb2.RequestApplySnapshotChunk.SerializeToString,
            tendermint_dot_abci_dot_types__pb2.ResponseApplySnapshotChunk.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2022-2023 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Chain constants"""

import os

from aea.configurations.constants import CONTRACTS, PACKAGES
from aea.configurations.data_types import PublicId

from autonomy import AUTONOMY_DIR
from autonomy.constants import VALORY
from autonomy.data import DATA_DIR


EVENT_VERIFICATION_TIMEOUT = 60.0

CONTRACTS_DIR_FRAMEWORK = DATA_DIR / CONTRACTS
CONTRACTS_DIR_LOCAL = (
    AUTONOMY_DIR.parent / PACKAGES / VALORY / CONTRACTS
)  # use from an editable/local installation


# contract addresses from `valory/autonolas-registries` image
COMPONENT_REGISTRY_ADDRESS_LOCAL = "0x5FbDB2315678afecb367f032d93F642f64180aa3"
AGENT_REGISTRY_ADDRESS_LOCAL = "0xe7f1725E7734CE288F8367e1Bb143E90bb3F0512"
REGISTRIES_MANAGER_ADDRESS_LOCAL = "0x9fE46736679d2D9a65F0992F2272dE9f3c7fa6e0"
SERVICE_REGISTRY_ADDRESS_LOCAL = "0x998abeb3E57409262aE5b751f60747921B33613E"
SERVICE_MANAGER_ADDRESS_LOCAL = "0x4c5859f0F772848b2D91F1D83E2Fe57935348029"
GNOSIS_SAFE_MULTISIG_ADDRESS_LOCAL = "0x0E801D84Fa97b50751Dbf25036d067dCf18858bF"


# contract addresses from `https://github.com/valory-xyz/autonolas-registries/blob/main/docs/mainnet_addresses.json` file
COMPONENT_REGISTRY_ADDRESS_ETHEREUM = "0x15bd56669F57192a97dF41A2aa8f4403e9491776"
AGENT_REGISTRY_ADDRESS_ETHEREUM = "0x2F1f7D38e4772884b88f3eCd8B6b9faCdC319112"
REGISTRIES_MANAGER_ADDRESS_ETHEREUM = "0x9eC9156dEF5C613B2a7D4c46C383F9B58DfcD6fE"
SERVICE_MANAGER_ADDRESS_ETHEREUM = "0x2EA682121f815FBcF86EA3F3CaFdd5d67F2dB143"
SERVICE_REGISTRY_ADDRESS_ETHEREUM = "0x48b6af7B12C71f09e2fC8aF4855De4Ff54e775cA"
GNOSIS_SAFE_MULTISIG_ADDRESS_ETHEREUM = "0x1c2cD884127b080F940b7546c1e9aaf525b1FA55"

# contract addresses from `https://raw.githubusercontent.com/valory-xyz/autonolas-registries/main/scripts/deployment/globals_goerli.json`
COMPONENT_REGISTRY_ADDRESS_GOERLI = "0x7Fd1F4b764fA41d19fe3f63C85d12bf64d2bbf68"
AGENT_REGISTRY_ADDRESS_GOERLI = "0xEB5638eefE289691EcE01943f768EDBF96258a80"
REGISTRIES_MANAGER_ADDRESS_GOERLI = "0x10c5525F77F13b28f42c5626240c001c2D57CAd4"
SERVICE_MANAGER_ADDRESS_GOERLI = "0x1d333b46dB6e8FFd271b6C2D2B254868BD9A2dbd"
SERVICE_REGISTRY_ADDRESS_GOERLI = "0x1cEe30D08943EB58EFF84DD1AB44a6ee6FEff63a"
GNOSIS_SAFE_MULTISIG_ADDRESS_GOERLI = "0x65dD51b02049ad1B6FF7fa9Ea3322E1D2CAb1176"

# contract addressed for a custom chain
COMPONENT_REGISTRY_ADDRESS_CUSTOM = os.environ.get("CUSTOM_COMPONENT_REGISTRY_ADDRESS")
AGENT_REGISTRY_ADDRESS_CUSTOM = os.environ.get("CUSTOM_AGENT_REGISTRY_ADDRESS")
REGISTRIES_MANAGER_ADDRESS_CUSTOM = os.environ.get("CUSTOM_REGISTRIES_MANAGER_ADDRESS")
SERVICE_MANAGER_ADDRESS_CUSTOM = os.environ.get("CUSTOM_SERVICE_MANAGER_ADDRESS")
SERVICE_REGISTRY_ADDRESS_CUSTOM = os.environ.get("CUSTOM_SERVICE_REGISTRY_ADDRESS")
GNOSIS_SAFE_MULTISIG_ADDRESS_CUSTOM = os.environ.get(
    "CUSTOM_GNOSIS_SAFE_MULTISIG_ADDRESS"
)

# Contract PublicIds
COMPONENT_REGISTRY_CONTRACT = PublicId.from_str("valory/component_registry")
AGENT_REGISTRY_CONTRACT = PublicId.from_str("valory/agent_registry")
REGISTRIES_MANAGER_CONTRACT = PublicId.from_str("valory/registries_manager")
SERVICE_MANAGER_CONTRACT = PublicId.from_str("valory/service_manager")
SERVICE_REGISTRY_CONTRACT = PublicId.from_str("valory/service_registry")
GNOSIS_SAFE_MULTISIG_CONTRACT = PublicId.from_str("valory/gnosis_safe_proxy_factory")



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2021-2023 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Test the dialogues.py module of the skill."""

# pylint: skip-file

from enum import Enum
from typing import Type, cast
from unittest.mock import MagicMock

import pytest
from aea.protocols.dialogue.base import Dialogues
from aea.skills.base import Model

from packages.valory.connections.ipfs.connection import PUBLIC_ID as IPFS_CONNECTION_ID
from packages.valory.protocols.ipfs import IpfsMessage
from packages.valory.skills.abstract_round_abci.dialogues import (
    AbciDialogue,
    AbciDialogues,
    ContractApiDialogue,
    ContractApiDialogues,
    HttpDialogue,
    HttpDialogues,
    IpfsDialogues,
    LedgerApiDialogue,
    LedgerApiDialogues,
    SigningDialogue,
    SigningDialogues,
    TendermintDialogue,
    TendermintDialogues,
)


@pytest.mark.parametrize(
    "dialogues_cls,expected_role_from_first_message",
    [
        (AbciDialogues, AbciDialogue.Role.CLIENT),
        (HttpDialogues, HttpDialogue.Role.CLIENT),
        (SigningDialogues, SigningDialogue.Role.SKILL),
        (LedgerApiDialogues, LedgerApiDialogue.Role.AGENT),
        (ContractApiDialogues, ContractApiDialogue.Role.AGENT),
        (TendermintDialogues, TendermintDialogue.Role.AGENT),
    ],
)
def test_dialogues_creation(
    dialogues_cls: Type[Model], expected_role_from_first_message: Enum
) -> None:
    """Test XDialogues creations."""
    dialogues = cast(Dialogues, dialogues_cls(name="", skill_context=MagicMock()))
    assert expected_role_from_first_message == dialogues._role_from_first_message(
        MagicMock(), MagicMock()
    )


def test_ledger_api_dialogue() -> None:
    """Test 'LedgerApiDialogue' creation."""
    dialogue = LedgerApiDialogue(MagicMock(), "", MagicMock())
    with pytest.raises(ValueError, match="Terms not set!"):
        dialogue.terms

    expected_terms = MagicMock()
    dialogue.terms = expected_terms
    assert expected_terms == dialogue.terms


def test_contract_api_dialogue() -> None:
    """Test 'ContractApiDialogue' creation."""
    dialogue = ContractApiDialogue(MagicMock(), "", MagicMock())
    with pytest.raises(ValueError, match="Terms not set!"):
        dialogue.terms

    expected_terms = MagicMock()
    dialogue.terms = expected_terms
    assert expected_terms == dialogue.terms


def test_ipfs_dialogue() -> None:
    """Test 'IpfsDialogues' creation."""
    dialogues = IpfsDialogues(name="", skill_context=MagicMock())
    dialogues.create(
        counterparty=str(IPFS_CONNECTION_ID),
        performative=IpfsMessage.Performative.GET_FILES,
    )



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2022-2023 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Chain interaction base classes."""

from enum import Enum
from typing import Optional

from aea.configurations.data_types import PublicId
from aea.contracts.base import Contract

from autonomy.chain.constants import (
    AGENT_REGISTRY_CONTRACT,
    COMPONENT_REGISTRY_CONTRACT,
    CONTRACTS_DIR_FRAMEWORK,
    CONTRACTS_DIR_LOCAL,
    REGISTRIES_MANAGER_CONTRACT,
    SERVICE_MANAGER_CONTRACT,
    SERVICE_REGISTRY_CONTRACT,
)


class UnitType(Enum):
    """
    Unit type

    Same as https://github.com/valory-xyz/autonolas-registries/blob/v1.0.1/contracts/interfaces/IRegistry.sol#L6-L9
    """

    COMPONENT = 0
    AGENT = 1


class ServiceState(Enum):
    """
    Service state

    Same as https://github.com/valory-xyz/autonolas-registries/blob/v1.0.1/contracts/ServiceRegistry.sol#L41-L48
    """

    NON_EXISTENT = 0
    PRE_REGISTRATION = 1
    ACTIVE_REGISTRATION = 2
    FINISHED_REGISTRATION = 3
    DEPLOYED = 4
    TERMINATED_BONDED = 5


class RegistryContracts:
    """On chain registry contracts helper"""

    _registries_manager: Optional[Contract] = None
    _service_manager: Optional[Contract] = None
    _component_registry: Optional[Contract] = None
    _agent_registry: Optional[Contract] = None
    _service_registry: Optional[Contract] = None

    @staticmethod
    def get_contract(public_id: PublicId) -> Contract:
        """Load contract for given public id."""

        # check if a local package is available
        contract_dir = CONTRACTS_DIR_LOCAL / public_id.name
        if contract_dir.exists():
            return Contract.from_dir(directory=contract_dir)

        # if local package is not available use one from the data directory
        contract_dir = CONTRACTS_DIR_FRAMEWORK / public_id.name
        if not contract_dir.exists():
            raise FileNotFoundError(
                f"Contract package {public_id} not found in the open-autonomy installation, "
                "please reinstall the package"
            )
        return Contract.from_dir(directory=contract_dir)

    @property
    def registries_manager(
        self,
    ) -> Contract:
        """Returns an instance of the registries manager contract."""
        if self._registries_manager is None:
            self._registries_manager = self.get_contract(
                public_id=REGISTRIES_MANAGER_CONTRACT
            )

        return self._registries_manager

    @property
    def service_manager(
        self,
    ) -> Contract:
        """Returns an instance of the registries manager contract."""
        if self._service_manager is None:
            self._service_manager = self.get_contract(
                public_id=SERVICE_MANAGER_CONTRACT
            )

        return self._service_manager

    @property
    def component_registry(
        self,
    ) -> Contract:
        """Returns an instance of the registries manager contract."""
        if self._component_registry is None:
            self._component_registry = self.get_contract(
                public_id=COMPONENT_REGISTRY_CONTRACT
            )

        return self._component_registry

    @property
    def agent_registry(
        self,
    ) -> Contract:
        """Returns an instance of the registries manager contract."""
        if self._agent_registry is None:
            self._agent_registry = self.get_contract(
                public_id=AGENT_REGISTRY_CONTRACT,
            )

        return self._agent_registry

    @property
    def service_registry(
        self,
    ) -> Contract:
        """Returns an instance of the registries manager contract."""
        if self._service_registry is None:
            self._service_registry = self.get_contract(
                public_id=SERVICE_REGISTRY_CONTRACT,
            )

        return self._service_registry


registry_contracts = RegistryContracts()



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2022 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Tests to ensure implementation is on par with ABCI spec"""

# pylint: skip-file

import logging
import time
from pathlib import Path
from typing import Any, Dict, Set

import requests
from aea.protocols.generator.common import (
    SPECIFICATION_COMPOSITIONAL_TYPES,
    _get_sub_types_of_compositional_types,
)

from packages.valory.connections import abci as tendermint_abci
from packages.valory.connections.abci.tests.helper import (
    camel_to_snake,
    compare_trees,
    create_aea_abci_type_tree,
    decode,
    encode,
    get_protocol_readme_spec,
    get_tender_type_tree,
    get_tendermint_content,
    init_aea_abci_messages,
    init_tendermint_messages,
    init_type_tree_primitives,
    replace_keys,
)


Node = Dict[str, Any]

# constants & utility functions
ENCODING = "utf-8"
VERSION = "v0.34.19"
REPO_PATH = Path(*tendermint_abci.__package__.split(".")).absolute()
PROTO_FILES = list((REPO_PATH / "protos" / "tendermint").glob("**/*.proto"))
URL_PREFIX = f"https://raw.githubusercontent.com/tendermint/tendermint/{VERSION}/proto/tendermint/"

# to ensure primitives are not initialized to empty default values
NON_DEFAULT_PRIMITIVES = {str: "sss", bytes: b"bbb", int: 123, float: 3.14, bool: True}
REPEATED_FIELD_SIZE = 3
USE_NON_ZERO_ENUM: bool = True


# tests
def test_local_types_file_matches_github(request_attempts: int = 3) -> None:
    """Test local file containing ABCI spec matches Tendermint GitHub"""

    different = []
    for file in PROTO_FILES:
        url = URL_PREFIX + "/".join(file.parts[-2:])
        response, i = requests.get(url), 0
        while response.status_code != 200 and i < request_attempts:
            time.sleep(0.1)
            response, i = requests.get(url), i + 1
        if response.status_code != 200:
            log_msg = "Failed to retrieve Tendermint proto types from Github"
            status_code, reason = response.status_code, response.reason
            raise requests.HTTPError(f"{log_msg}: {status_code} ({reason})")
        github_data = response.text
        local_data = file.read_text(encoding=ENCODING)
        if not github_data == local_data:
            different.append([file, url])

    if different:
        logging.error("\n".join(" =/= ".join(map(str, x)) for x in different))
    assert not bool(different)


def test_all_custom_types_used() -> None:
    """
    Test if all custom types are used in speech acts.

    By asserting their usage in the speech acts we can delegate
    the verification of their implementation and translation to
    another test that addresses this (test_aea_to_tendermint).
    """

    aea_protocol, custom_types, _ = get_protocol_readme_spec()
    speech_acts = aea_protocol["speech_acts"]
    defined_types = {v for vals in speech_acts.values() for v in vals.values()}

    custom_in_speech: Set[str] = set()
    for d_type in defined_types:
        if any(map(d_type.startswith, SPECIFICATION_COMPOSITIONAL_TYPES)):
            subfields = _get_sub_types_of_compositional_types(d_type)
            custom_in_speech.update(s[3:] for s in subfields if s.startswith("ct:"))
        elif d_type.startswith("ct:"):
            custom_in_speech.add(d_type[3:])

    assert custom_in_speech == {s[3:] for s in custom_types}


def test_defined_dialogues_match_abci_spec() -> None:
    """
    Test defined dialogues match ABCI spec.

    It verifies solely that request response pairs match:
      - AEA requests match Tendermint requests
      - AEA responses match Tendermint responses
      - That all requests have a matching response
      - That all request, response and the exception are covered
    """

    *_, dialogues = get_protocol_readme_spec()
    tender_type_tree = get_tender_type_tree()

    # expected
    request_oneof = tender_type_tree["Request"][-1]
    request_keys = {camel_to_snake(cls.__name__) for cls, _ in request_oneof.values()}
    response_oneof = tender_type_tree["Response"][-1]
    response_keys = {camel_to_snake(cls.__name__) for cls, _ in response_oneof.values()}

    # defined
    initiation = dialogues["initiation"]
    reply = dialogues["reply"]
    termination = dialogues["termination"]

    # initiation
    assert not request_keys.difference(initiation)
    assert set(initiation).difference(request_keys) == {"dummy"}

    # reply
    missing_response, alt = set(), "response_exception"
    for key in request_keys:
        if not set(reply[key]) == {key.replace("request", "response"), alt}:
            missing_response.add(key)
    assert not missing_response
    assert not any(reply[key] for key in response_keys)

    # termination
    assert not response_keys.difference(termination)
    assert set(termination).difference(response_keys) == {"dummy"}


def test_aea_to_tendermint() -> None:
    """
    Test translation from AEA-native to Tendermint-native ABCI protocol.

    "repeated" fields are returned as list in python,
    but must be passed as tuples to Tendermint protobuf.
    """

    aea_protocol, *_ = get_protocol_readme_spec()
    speech_acts = aea_protocol["speech_acts"]

    # 1. create type tree from speech acts
    type_tree = create_aea_abci_type_tree(speech_acts)
    type_tree.pop("dummy")  # TODO: known oddity on our side

    # 2. initialize primitives
    init_tree = init_type_tree_primitives(type_tree)

    # 3. create AEA-native ABCI protocol messages
    abci_messages = init_aea_abci_messages(type_tree, init_tree)

    # 4. encode to Tendermint-native ABCI protocol
    #    NOTE: request not implemented in encoder
    encoded = {k: encode(v) for k, v in abci_messages.items()}
    translated = {k: v for k, v in encoded.items() if v}
    untranslated = set(encoded).difference(translated)
    assert all(k.startswith("request") for k in untranslated)

    # 5. create Tendermint message content tree
    tender_tree = {k: get_tendermint_content(v) for k, v in translated.items()}

    # 6. translate expected differences in attribute / field naming
    #    these are known difference introduced by translating between protocols
    param_keys_trans = {k: f"{k}_params" for k in ["evidence", "validator", "version"]}
    tendermint_to_aea = dict(
        response_info={"data": "info_data"},
        response_query={"proof_ops": {"ops": "proof_ops"}},
        response_init_chain={"consensus_params": param_keys_trans},
        response_end_block={"consensus_param_updates": param_keys_trans},
    )
    replace_keys(tender_tree, tendermint_to_aea)

    # 7. compare AEA-native message initialization with the information
    #    retrieved from the Tendermint Response after translation
    shared = set(type_tree).intersection(tender_tree)
    assert len(shared) == 16  # expected number of matches
    for k in shared:
        init_node, tender_node = init_tree[k], tender_tree[k]
        compare_trees(init_node, tender_node)


def test_tendermint_decoding() -> None:
    """Test Tendermint ABCI message decoding"""

    # 1. create tendermint type tree
    tender_type_tree = get_tender_type_tree()

    # 2. initialize messages
    messages = init_tendermint_messages(tender_type_tree)

    # 3. translate to AEA-native ABCI Messages
    decoded = list(map(decode, messages))
    assert len(decoded) == 15  # expected number of matches



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2021-2023 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""This module contains the behaviours for the 'registration_abci' skill."""

import datetime
import json
from abc import ABC
from enum import Enum
from typing import Any, Dict, Generator, Optional, Set, Type, cast

from aea.mail.base import EnvelopeContext

from packages.valory.connections.p2p_libp2p_client.connection import (
    PUBLIC_ID as P2P_LIBP2P_CLIENT_PUBLIC_ID,
)
from packages.valory.contracts.service_registry.contract import ServiceRegistryContract
from packages.valory.protocols.contract_api import ContractApiMessage
from packages.valory.protocols.http import HttpMessage
from packages.valory.protocols.tendermint import TendermintMessage
from packages.valory.skills.abstract_round_abci.base import ABCIAppInternalError
from packages.valory.skills.abstract_round_abci.behaviour_utils import TimeoutException
from packages.valory.skills.abstract_round_abci.behaviours import (
    AbstractRoundBehaviour,
    BaseBehaviour,
)
from packages.valory.skills.abstract_round_abci.utils import parse_tendermint_p2p_url
from packages.valory.skills.registration_abci.dialogues import TendermintDialogues
from packages.valory.skills.registration_abci.models import SharedState
from packages.valory.skills.registration_abci.payloads import RegistrationPayload
from packages.valory.skills.registration_abci.rounds import (
    AgentRegistrationAbciApp,
    RegistrationRound,
    RegistrationStartupRound,
)


NODE = "node_{address}"
WAIT_FOR_BLOCK_TIMEOUT = 60.0  # 1 minute


class RegistrationBaseBehaviour(BaseBehaviour, ABC):
    """Agent registration to the FSM App."""

    def async_act(self) -> Generator:
        """
        Do the action.

        Steps:
        - Build a registration transaction.
        - Send the transaction and wait for it to be mined.
        - Wait until ABCI application transitions to the next round.
        - Go to the next behaviour (set done event).
        """

        with self.context.benchmark_tool.measure(self.behaviour_id).local():
            serialized_db = self.synchronized_data.db.serialize()
            payload = RegistrationPayload(
                self.context.agent_address, initialisation=serialized_db
            )

        with self.context.benchmark_tool.measure(self.behaviour_id).consensus():
            yield from self.send_a2a_transaction(payload)
            yield from self.wait_until_round_end()

        self.set_done()


class RegistrationStartupBehaviour(RegistrationBaseBehaviour):
    """Agent registration to the FSM App."""

    ENCODING: str = "utf-8"
    matching_round = RegistrationStartupRound
    local_tendermint_params: Dict[str, Any] = {}
    updated_genesis_data: Dict[str, Any] = {}
    collection_complete: bool = False

    @property
    def initial_tm_configs(self) -> Dict[str, Dict[str, Any]]:
        """A mapping of the other agents' addresses to their initial Tendermint configuration."""
        return self.context.state.initial_tm_configs

    @initial_tm_configs.setter
    def initial_tm_configs(self, configs: Dict[str, Dict[str, Any]]) -> None:
        """A mapping of the other agents' addresses to their initial Tendermint configuration."""
        self.context.state.initial_tm_configs = configs

    class LogMessages(Enum):
        """Log messages used in RegistrationStartupBehaviour"""

        config_sharing = "Sharing Tendermint config on start-up?"
        # request personal tendermint configuration
        request_personal = "Request validator config from personal Tendermint node"
        response_personal = "Response validator config from personal Tendermint node"
        failed_personal = "Failed validator config from personal Tendermint node"
        # verify deployment on-chain contract
        request_verification = "Request service registry contract verification"
        response_verification = "Response service registry contract verification"
        failed_verification = "Failed service registry contract verification"
        # request service info from on-chain contract
        request_service_info = "Request on-chain service info"
        response_service_info = "Response on-chain service info"
        failed_service_info = "Failed on-chain service info"
        # request tendermint configuration other agents
        request_others = "Request Tendermint config info from other agents"
        collection_complete = "Completed collecting Tendermint configuration responses"
        # update personal tendermint node config
        request_update = "Request update Tendermint node configuration"
        response_update = "Response update Tendermint node configuration"
        failed_update = "Failed update Tendermint node configuration"
        # exceptions
        no_contract_address = "Service registry contract address not provided"
        no_on_chain_service_id = "On-chain service id not provided"
        contract_incorrect = "Service registry contract not correctly deployed"
        no_agents_registered = "No agents registered on-chain"
        self_not_registered = "This agent is not registered on-chain"

        def __str__(self) -> str:
            """For ease of use in formatted string literals"""
            return self.value

    @property
    def tendermint_parameter_url(self) -> str:
        """Tendermint URL for obtaining and updating parameters"""
        return f"{self.params.tendermint_com_url}/params"

    def _decode_result(
        self, message: HttpMessage, error_log: LogMessages
    ) -> Optional[Dict[str, Any]]:
        """Decode a http message's body.

        :param message: the http message.
        :param error_log: a log to prefix potential errors with.
        :return: the message's body, as a dictionary
        """
        try:
            response = json.loads(message.body.decode())
        except json.JSONDecodeError as error:
            self.context.logger.error(f"{error_log}: {error}")
            return None

        if not response["status"]:  # pragma: no cover
            self.context.logger.error(f"{error_log}: {response['error']}")
            return None

        return response

    def is_correct_contract(
        self, service_registry_address: str
    ) -> Generator[None, None, bool]:
        """Contract deployment verification."""

        self.context.logger.info(self.LogMessages.request_verification)

        performative = ContractApiMessage.Performative.GET_STATE
        kwargs = dict(
            performative=performative,
            contract_address=service_registry_address,
            contract_id=str(ServiceRegistryContract.contract_id),
            contract_callable="verify_contract",
        )
        contract_api_response = yield from self.get_contract_api_response(**kwargs)  # type: ignore
        if (
            contract_api_response.performative
            is not ContractApiMessage.Performative.STATE
        ):
            verified = False
            log_method = self.context.logger.error
            log_message = f"{self.LogMessages.failed_verification} ({kwargs}): {contract_api_response}"
        else:
            verified = cast(bool, contract_api_response.state.body["verified"])
            log_method = self.context.logger.info
            log_message = f"{self.LogMessages.response_verification}: {verified}"

        log_method(log_message)
        return verified

    def get_agent_instances(
        self, service_registry_address: str, on_chain_service_id: int
    ) -> Generator[None, None, Dict[str, Any]]:
        """Get service info available on-chain"""

        log_message = self.LogMessages.request_service_info
        self.context.logger.info(f"{log_message}")

        performative = ContractApiMessage.Performative.GET_STATE
        kwargs = dict(
            performative=performative,
            contract_address=service_registry_address,
            contract_id=str(ServiceRegistryContract.contract_id),
            contract_callable="get_agent_instances",
            service_id=on_chain_service_id,
        )
        contract_api_response = yield from self.get_contract_api_response(**kwargs)  # type: ignore
        if contract_api_response.performative != ContractApiMessage.Performative.STATE:
            log_message = self.LogMessages.failed_service_info
            self.context.logger.error(
                f"{log_message} ({kwargs}): {contract_api_response}"
            )
            return {}

        log_message = self.LogMessages.response_service_info
        self.context.logger.info(f"{log_message}: {contract_api_response}")
        return cast(dict, contract_api_response.state.body)

    def get_addresses(self) -> Generator:  # pylint: disable=too-many-return-statements
        """Get addresses of agents registered for the service"""

        service_registry_address = self.params.service_registry_address
        if service_registry_address is None:
            log_message = self.LogMessages.no_contract_address.value
            self.context.logger.error(log_message)
            return False

        correctly_deployed = yield from self.is_correct_contract(
            service_registry_address
        )
        if not correctly_deployed:
            return False

        on_chain_service_id = self.params.on_chain_service_id
        if on_chain_service_id is None:
            log_message = self.LogMessages.no_on_chain_service_id.value
            self.context.logger.error(log_message)
            return False

        service_info = yield from self.get_agent_instances(
            service_registry_address, on_chain_service_id
        )
        if not service_info:
            return False

        registered_addresses = set(service_info["agentInstances"])
        if not registered_addresses:
            log_message = self.LogMessages.no_agents_registered.value
            self.context.logger.error(f"{log_message}: {service_info}")
            return False

        my_address = self.context.agent_address
        if my_address not in registered_addresses:
            log_message = f"{self.LogMessages.self_not_registered} ({my_address})"
            self.context.logger.error(f"{log_message}: {registered_addresses}")
            return False

        # put service info in the shared state for p2p message handler
        info: Dict[str, Dict[str, str]] = {i: {} for i in registered_addresses}
        tm_host, tm_port = parse_tendermint_p2p_url(url=self.params.tendermint_p2p_url)
        validator_config = dict(
            hostname=tm_host,
            p2p_port=tm_port,
            address=self.local_tendermint_params["address"],
            pub_key=self.local_tendermint_params["pub_key"],
            peer_id=self.local_tendermint_params["peer_id"],
        )
        info[self.context.agent_address] = validator_config
        self.initial_tm_configs = info
        log_message = self.LogMessages.response_service_info.value
        self.context.logger.info(f"{log_message}: {info}")
        return True

    def get_tendermint_configuration(self) -> Generator[None, None, bool]:
        """Make HTTP GET request to obtain agent's local Tendermint node parameters"""

        url = self.tendermint_parameter_url
        log_message = self.LogMessages.request_personal
        self.context.logger.info(f"{log_message}: {url}")

        result = yield from self.get_http_response(method="GET", url=url)
        response = self._decode_result(result, self.LogMessages.failed_personal)
        if response is None:
            return False

        self.local_tendermint_params = response["params"]
        log_message = self.LogMessages.response_personal
        self.context.logger.info(f"{log_message}: {response}")
        return True

    def request_tendermint_info(self) -> Generator[None, None, bool]:
        """Request Tendermint info from other agents"""

        still_missing = {k for k, v in self.initial_tm_configs.items() if not v} - {
            self.context.agent_address
        }
        log_message = self.LogMessages.request_others
        self.context.logger.info(f"{log_message}: {still_missing}")

        for address in still_missing:
            dialogues = cast(TendermintDialogues, self.context.tendermint_dialogues)
            performative = TendermintMessage.Performative.GET_GENESIS_INFO
            message, _ = dialogues.create(
                counterparty=address, performative=performative
            )
            message = cast(TendermintMessage, message)
            context = EnvelopeContext(connection_id=P2P_LIBP2P_CLIENT_PUBLIC_ID)
            self.context.outbox.put_message(message=message, context=context)
        # we wait for the messages that were put in the outbox.
        yield from self.sleep(self.params.sleep_time)

        if all(self.initial_tm_configs.values()):
            log_message = self.LogMessages.collection_complete
            self.context.logger.info(f"{log_message}: {self.initial_tm_configs}")
            self.collection_complete = True
        return self.collection_complete

    def format_genesis_data(
        self,
        collected_agent_info: Dict[str, Any],
    ) -> Dict[str, Any]:
        """Format collected agent info for genesis update"""

        validators = []
        for address, validator_config in collected_agent_info.items():
            validator = dict(
                hostname=validator_config["hostname"],
                p2p_port=validator_config["p2p_port"],
                address=validator_config["address"],
                pub_key=validator_config["pub_key"],
                peer_id=validator_config["peer_id"],
                power=self.params.genesis_config.voting_power,
                name=NODE.format(address=address[2:]),  # skip 0x part
            )
            validators.append(validator)

        genesis_data = dict(
            validators=validators,
            genesis_config=self.params.genesis_config.to_json(),
            external_address=self.params.tendermint_p2p_url,
        )
        return genesis_data

    def request_update(self) -> Generator[None, None, bool]:
        """Make HTTP POST request to update agent's local Tendermint node"""

        url = self.tendermint_parameter_url
        genesis_data = self.format_genesis_data(self.initial_tm_configs)
        log_message = self.LogMessages.request_update
        self.context.logger.info(f"{log_message}: {genesis_data}")

        content = json.dumps(genesis_data).encode(self.ENCODING)
        result = yield from self.get_http_response(
            method="POST", url=url, content=content
        )
        response = self._decode_result(result, self.LogMessages.failed_update)
        if response is None:
            return False

        log_message = self.LogMessages.response_update
        self.context.logger.info(f"{log_message}: {response}")
        self.updated_genesis_data.update(genesis_data)
        return True

    def wait_for_block(self, timeout: float) -> Generator[None, None, bool]:
        """Wait for a block to be received in the specified timeout."""
        # every agent will finish with the reset at a different time
        # hence the following will be different for all agents
        start_time = datetime.datetime.now()

        def received_block() -> bool:
            """Check whether we have received a block after "start_time"."""
            try:
                shared_state = cast(SharedState, self.context.state)
                last_timestamp = shared_state.round_sequence.last_timestamp
                if last_timestamp > start_time:
                    return True
                return False
            except ABCIAppInternalError:
                # this can happen if we haven't received a block yet
                return False

        try:
            yield from self.wait_for_condition(
                condition=received_block, timeout=timeout
            )
            # if the `wait_for_condition` finish without an exception,
            # it means that the condition has been satisfied on time
            return True
        except TimeoutException:
            # the agent wasn't able to receive blocks in the given amount of time (timeout)
            return False

    def async_act(self) -> Generator:  # pylint: disable=too-many-return-statements
        """
        Do the action.

        Steps:
        1. Collect personal Tendermint configuration
        2. Make Service Registry contract call to retrieve addresses
           of the other agents registered on-chain for the service.
        3. Request Tendermint configuration from registered agents.
           This is done over the Agent Communication Network using
           the p2p_libp2p_client connection.
        4. Update Tendermint configuration via genesis.json with the
           information of the other validators (agents).
        5. Restart Tendermint to establish the validator network.
        """

        exchange_config = self.params.share_tm_config_on_startup
        log_message = self.LogMessages.config_sharing.value
        self.context.logger.info(f"{log_message}: {exchange_config}")

        if not exchange_config:
            yield from super().async_act()
            return

        self.context.logger.info(f"My address: {self.context.agent_address}")

        # collect personal Tendermint configuration
        if not self.local_tendermint_params:
            successful = yield from self.get_tendermint_configuration()
            if not successful:
                yield from self.sleep(self.params.sleep_time)
                return

        # if the agent doesn't have it's tm config info set, then make service registry contract call
        # to get the rest of the agents, so we can get their tm config info later
        info = self.initial_tm_configs.get(self.context.agent_address, None)
        if info is None:
            successful = yield from self.get_addresses()
            if not successful:
                yield from self.sleep(self.params.sleep_time)
                return

        # collect Tendermint config information from other agents
        if not self.collection_complete:
            successful = yield from self.request_tendermint_info()
            if not successful:
                yield from self.sleep(self.params.sleep_time)
                return

        # update Tendermint configuration
        if not self.updated_genesis_data:
            successful = yield from self.request_update()
            if not successful:
                yield from self.sleep(self.params.sleep_time)
                return

        # restart Tendermint with updated configuration
        successful = yield from self.reset_tendermint_with_wait(on_startup=True)
        if not successful:
            yield from self.sleep(self.params.sleep_time)
            return

        # the reset has gone through, and at this point tendermint should start
        # sending blocks to the agent. However, that might take a while, since
        # we rely on 2/3 of the voting power to be active in order for block production
        # to begin. In other words, we wait for >=2/3 of the agents to become active.
        successful = yield from self.wait_for_block(timeout=WAIT_FOR_BLOCK_TIMEOUT)
        if not successful:
            yield from self.sleep(self.params.sleep_time)
            return

        yield from super().async_act()


class RegistrationBehaviour(RegistrationBaseBehaviour):
    """Agent registration to the FSM App."""

    matching_round = RegistrationRound


class AgentRegistrationRoundBehaviour(AbstractRoundBehaviour):
    """This behaviour manages the consensus stages for the registration."""

    initial_behaviour_cls = RegistrationStartupBehaviour
    abci_app_cls = AgentRegistrationAbciApp
    behaviours: Set[Type[BaseBehaviour]] = {
        RegistrationBehaviour,  # type: ignore
        RegistrationStartupBehaviour,  # type: ignore
    }



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2022-2023 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Test check-handlers command."""

import os
import shutil
from pathlib import Path
from typing import Tuple
from unittest import mock

import pytest
from aea.configurations.constants import DEFAULT_SKILL_CONFIG_FILE

from autonomy.analyse.handlers import check_handlers

from tests.conftest import ROOT_DIR
from tests.test_autonomy.test_cli.base import BaseCliTest


COMMON_HANDLERS = (
    "abci",
    "http",
    "contract_api",
    "ledger_api",
    "signing",
)

IGNORE_SKILLS = (
    "abstract_abci",
    "counter",
    "counter_client",
    "hello_world_abci",
)


class TestCheckHandlers(BaseCliTest):
    """Test check-handlers command."""

    cli_options: Tuple[str, ...] = ("analyse", "handlers")
    config_file: Path
    config_file_temp: Path

    def setup(self) -> None:
        """Setup."""
        super().setup()

        shutil.copytree(ROOT_DIR / "packages", self.t / "packages")
        os.chdir(self.t)

        options = []
        for h in COMMON_HANDLERS:
            options += ["-h", h]

        for i in IGNORE_SKILLS:
            options += ["-i", i]

        self.cli_options = (*self.cli_options, *options)

    def test_check_handlers(
        self,
    ) -> None:
        """Run tests."""

        result = self.run_cli()
        assert result.exit_code == 0, result.output

        for yaml_file in sorted(
            Path(self.t / "packages").glob(f"*/*/*/{DEFAULT_SKILL_CONFIG_FILE}")
        ):
            if yaml_file.parent.name in IGNORE_SKILLS:
                assert f"Skipping {yaml_file.parent.name}" in result.output
            else:
                assert f"Checking {yaml_file.parent.name}" in result.output

    def test_check_handlers_fail(
        self,
    ) -> None:
        """Test check-handlers command fail."""
        result = self.run_cli(("-h", "dummy"))

        assert result.exit_code == 1, result.output
        assert "Common handler 'dummy' is not defined in" in result.output

    def test_check_handlers_missing_handler(
        self,
    ) -> None:
        """Test check-handlers command missing handler."""

        # Since the CLI catches exceptions, to test for raises we need to call check_handlers directly
        with pytest.raises(
            ValueError,
            match=r"Handler ABCIHandler declared in .* is missing from .*",
        ):
            # Mock dir() so module_attributes is = []
            with mock.patch("autonomy.analyse.handlers.dir", return_value=[]):
                # Mock Path.relative_to() and return any valid module so import_module does not fail
                with mock.patch(
                    "autonomy.analyse.handlers.load_handler_module_from_skill_path",
                    return_value=None,
                ):
                    check_handlers(
                        Path(
                            self.t,
                            "packages",
                            "valory",
                            "skills",
                            "abstract_abci",
                            "skill.yaml",
                        ),
                        [],
                    )


def test_check_handlers_missing_file() -> None:
    """Test check-handlers missing file."""

    # Since the CLI catches exceptions, to test for raises we need to call check_handlers directly
    with pytest.raises(FileNotFoundError, match="Handler file does not exist in file"):
        with mock.patch("pathlib.Path.relative_to", return_value=Path("dummy")):
            check_handlers(config_file=Path("file", "dummy"), common_handlers=[])



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2022-2023 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Tests for specs commands."""


import copy
import importlib
import os
import re
import shutil
from enum import Enum
from pathlib import Path
from typing import Tuple, cast
from unittest import mock

import pytest
from aea.configurations.constants import PACKAGES
from jsonschema.exceptions import ValidationError

from autonomy.analyse.abci.app_spec import (
    DFA,
    DFASpecificationError,
    FSMSpecificationLoader,
    check_unreferenced_events,
)

from tests.conftest import ROOT_DIR
from tests.test_autonomy.test_cli.base import BaseCliTest


class TestGenerateSpecs(BaseCliTest):
    """Test generate-app-specs"""

    cli_options: Tuple[str, ...] = ("analyse", "fsm-specs")

    dfa: DFA
    app_name: str
    skill_path: Path

    def setup(self) -> None:
        """Setup test method."""
        super().setup()

        self.app_name = "HelloWorldAbciApp"
        self.skill_path = Path(PACKAGES, "valory", "skills", "hello_world_abci")

        module_name = ".".join((*self.skill_path.parts, "rounds"))
        module = importlib.import_module(module_name)
        abci_app_cls = getattr(module, self.app_name)

        shutil.copytree(ROOT_DIR / PACKAGES, self.t / PACKAGES)

        self.dfa = DFA.abci_to_dfa(abci_app_cls)
        self.cli_options = (
            "--registry-path",
            str(self.t / PACKAGES),
            "analyse",
            "fsm-specs",
        )

        os.chdir(self.t)

    def get_expected_output(self, output_format: str) -> str:
        """Get expected output."""

        temp_file = self.t / "temp"
        FSMSpecificationLoader.dump(self.dfa, file=temp_file, spec_format=output_format)

        return temp_file.read_text(encoding="utf-8")

    def _run_test(self, output_format: str) -> None:
        """Run test for given output format type."""

        output_file = self.skill_path / cast(
            Path,
            FSMSpecificationLoader.OutputFormats.default_output_files.get(
                output_format
            ),
        )
        result = self.run_cli(
            (
                "--package",
                str(self.skill_path),
                f"--{output_format}",
                "--app-class",
                self.app_name,
                "--update",
            )
        )

        assert result.exit_code == 0, result.output
        assert output_file.read_text() == self.get_expected_output(output_format)

    def test_generate_yaml(
        self,
    ) -> None:
        """Run tests."""

        self._run_test(FSMSpecificationLoader.OutputFormats.YAML)

    def test_generate_json(
        self,
    ) -> None:
        """Run tests."""

        self._run_test(FSMSpecificationLoader.OutputFormats.JSON)

    def test_generate_mermaid(
        self,
    ) -> None:
        """Run tests."""

        self._run_test(FSMSpecificationLoader.OutputFormats.MERMAID)

    def test_failures(
        self,
    ) -> None:
        """Test failures."""

        with pytest.raises(ValueError, match="Unrecognized input format .exe"):
            FSMSpecificationLoader.dump(self.dfa, self.skill_path, ".exe")

        with pytest.raises(ValueError, match="Unrecognized input format .exe"):
            FSMSpecificationLoader.load(self.skill_path, ".exe")

        result = self.run_cli(
            (
                "--app-class",
                "SOME_CLASS_NAME",
                "--package",
                str(Path(*self.skill_path.parts, "dummy")),
                "--yaml",
            )
        )

        assert result.exit_code == 1, result.output
        assert (
            "Cannot find the rounds module or the composition module" in result.stdout
        ), result.output

        result = self.run_cli(
            (
                "--app-class",
                "SomeAppName",
                "--package",
                str(self.skill_path),
                "--yaml",
            )
        )

        assert result.exit_code == 1, result.output
        assert 'Class "SomeAppName" is not in' in result.stdout, result.output


class TestCheckSpecs(BaseCliTest):
    """Test `check-app-specs` command."""

    cli_options: Tuple[str, ...] = ("analyse", "fsm-specs")
    skill_path = Path(PACKAGES, "valory", "skills", "hello_world_abci")
    module_name = ".".join(skill_path.parts)
    app_name = "HelloWorldAbciApp"
    cls_name = ".".join([module_name, app_name])

    packages_dir: Path
    specification_path: Path

    def setup(self) -> None:
        """Setup class."""
        super().setup()

        self.packages_dir = self.t / PACKAGES
        shutil.copytree(ROOT_DIR / PACKAGES, self.packages_dir)
        self.specification_path = (
            self.t / self.skill_path.parent / "fsm_specification.yaml"
        )

        # make a copy of 'packages' in a subdirectory with depth > 1 from cwd
        subdirectory = self.t / Path("path", "to", "subdirectory")
        self.packages_dir_in_subdir = subdirectory / PACKAGES
        self.specification_path_in_subdir = (
            subdirectory / self.skill_path.parent / "fsm_specification.yaml"
        )
        shutil.copytree(self.packages_dir, self.packages_dir_in_subdir)

        self.specification_path = self.t / self.skill_path / "fsm_specification.yaml"
        os.chdir(self.t)

    def _corrupt_spec_file(
        self,
    ) -> None:
        """Corrupt spec file to fail the check."""
        content = self.specification_path.read_text()
        content = content.replace(
            "(SelectKeeperRound, ROUND_TIMEOUT): RegistrationRound\n", ""
        )
        self.specification_path.write_text(content)

    def test_one_pass(
        self,
    ) -> None:
        """Test with one class."""
        return_code, stdout, stderr = self.run_cli_subprocess(
            ("--app-class", self.app_name, "--package", str(self.skill_path))
        )

        assert return_code == 0, stderr
        assert "Check successful" in stdout

    def test_one_fail(
        self,
    ) -> None:
        """Test with one class failing."""
        self._corrupt_spec_file()
        return_code, stdout, stderr = self.run_cli_subprocess(
            ("--app-class", self.app_name, "--package", str(self.skill_path))
        )

        assert return_code == 1, stderr
        assert (
            "FSM Spec definition does not match in specification file and class definitions"
            in stderr
        )

    def test_check_all(
        self,
    ) -> None:
        """Test --check-all flag."""
        return_code, stdout, stderr = self.run_cli_subprocess(())

        assert return_code == 0, stderr
        assert "Done" in stdout

    def test_check_all_when_packages_is_not_in_working_dir(
        self,
    ) -> None:
        """Test --check-all flag when the packages directory is not in the working directory."""
        return_code, stdout, stderr = self.run_cli_subprocess(())

        assert return_code == 0
        assert "Checking all available packages" in stdout
        assert "Done" in stdout

    def test_check_fail_when_packages_dir_is_not_named_packages(
        self,
    ) -> None:
        """Test --check-all flag when the packages directory is not named 'packages'."""
        wrong_dir = self.t / "some-directory"
        wrong_dir.mkdir(exist_ok=True)

        self.cli_options = (
            "--registry-path",
            str(wrong_dir),
            "analyse",
            "fsm-specs",
        )
        return_code, _, stderr = self.run_cli_subprocess(())

        assert return_code == 1, stderr
        assert f"packages directory {wrong_dir} is not named '{PACKAGES}'" in stderr

    def test_check_all_fail(
        self,
    ) -> None:
        """Test --check-all flag."""
        self._corrupt_spec_file()
        return_code, stdout, stderr = self.run_cli_subprocess(())

        assert return_code == 1
        assert "Specifications check for following packages failed" in stderr

    def teardown(
        self,
    ) -> None:
        """Tear down the test."""
        super().teardown()


class TestDFA:
    """Test the DFA class."""

    good_dfa_kwargs = dict(
        label="DummyAbciApp",
        states={"StateARound", "StateBRound", "StateCRound"},
        default_start_state="StateARound",
        start_states={"StateARound"},
        final_states={"StateCRound"},
        alphabet_in={"event_a", "event_b", "event_c"},
        transition_func={
            ("StateARound", "event_b"): "StateBRound",
            ("StateBRound", "event_a"): "StateARound",
            ("StateBRound", "event_c"): "StateCRound",
        },
    )

    bad_dfa_kwargs = dict(
        label="DummyAbciApp",
        states={"StateARound", "StateBRound", "StateCRound", "unreachable_state"},
        default_start_state="state_other",
        start_states={"StateARound", "ExtraRound"},
        final_states={"StateARound", "StateCRound", "ExtraRound"},
        alphabet_in={"event_a", "event_b", "event_c", "other_extra_event"},
        transition_func={
            ("StateARound", "event_b"): "StateBRound",
            ("StateBRound", "event_a"): "StateARound",
            ("StateBRound", "event_c"): "StateCRound",
            ("ExtraRound", "extra_event"): "ExtraRound",
        },
    )

    def test_dfa(self) -> None:
        """Test DFA."""
        good_dfa = DFA(**self.good_dfa_kwargs)  # type: ignore

        assert not good_dfa.is_transition_func_total()
        assert good_dfa.get_transitions(["event_a"]) == ["StateARound", "StateARound"]
        assert good_dfa.get_transitions(["event_x"]) == ["StateARound"]
        assert isinstance(good_dfa.parse_transition_func(), dict)
        assert good_dfa.__eq__(None) == NotImplemented

        with pytest.raises(
            DFASpecificationError, match="DFA spec. object {} is not of type List."
        ):
            assert good_dfa._norep_list_to_set(dict())  # type: ignore

        with pytest.raises(
            DFASpecificationError,
            match=re.escape(
                "DFA spec. List ['value', 'value'] contains repeated values."
            ),
        ):
            assert good_dfa._norep_list_to_set(["value", "value"])

        with pytest.raises(
            DFASpecificationError,
            match="DFA spec. JSON file contains an invalid transition function key: .",
        ):
            assert good_dfa._str_to_tuple("")

        with pytest.raises(
            DFASpecificationError,
            match=re.escape(
                "DFA spec. JSON file contains an invalid transition function key: (a, )."
            ),
        ):
            assert good_dfa._str_to_tuple("(a, )")

        with pytest.raises(
            DFASpecificationError,
            match=re.escape(
                "DFA spec. JSON file contains an invalid transition function key: (, b)."
            ),
        ):
            assert good_dfa._str_to_tuple("(, b)")

        with pytest.raises(
            DFASpecificationError,
            match=re.escape(
                "DFA spec. JSON file contains an invalid transition function key: (, )."
            ),
        ):
            assert good_dfa._str_to_tuple("(, )")

        with pytest.raises(
            DFASpecificationError, match="DFA spec has the following issues"
        ):
            DFA(**self.bad_dfa_kwargs)  # type: ignore

    def test_load(self) -> None:
        """Test test_load"""

        json_spec = Path(
            ROOT_DIR,
            "tests",
            "data",
            "specs",
            "fsm_specification.json",
        )

        assert isinstance(
            DFA.load(json_spec, FSMSpecificationLoader.OutputFormats.JSON), DFA
        )

        with pytest.raises(ValueError):
            DFA.load(json_spec, "wrong_format")

    def test_load_empty(self) -> None:
        """Test test_load_empty"""

        json_spec = Path(
            ROOT_DIR,
            "tests",
            "data",
            "specs",
            "fsm_specification_empty.json",
        )

        with pytest.raises(ValidationError, match="is a required property"):
            DFA.load(json_spec, FSMSpecificationLoader.OutputFormats.JSON)

    def test_load_extra(self) -> None:
        """Test test_load_extra"""

        json_spec = Path(
            ROOT_DIR,
            "tests",
            "data",
            "specs",
            "fsm_specification_extra.json",
        )

        with pytest.raises(
            ValidationError,
            match=re.escape("Additional properties are not allowed"),
        ):
            DFA.load(json_spec, FSMSpecificationLoader.OutputFormats.JSON)

    def test_check_unreferenced_events(self) -> None:
        """Test check_unreferenced_events"""

        class MockABCIApp:
            """Mock ABCIApp class"""

            class Round:
                """Mock Round class"""

                __name__ = "round"

            class Event(Enum):
                """Mock Event class"""

                A = "A"
                B = "B"

            initial_round_cls = "initial_round_cls"
            transition_function = {
                Round: {
                    Event.A: "round_b",
                    Event.B: "round_b",
                },
            }
            event_to_timeout = {
                Event.A: 30.0,
            }

        with mock.patch("inspect.getmro", return_value=[]):
            strings = check_unreferenced_events(MockABCIApp)
            assert len(strings) > 0

    def test_name_verification(self) -> None:
        """Test `validate_naming_conventions`"""

        dfa_kwargs = copy.deepcopy(self.good_dfa_kwargs)
        dfa_kwargs["label"] = "abci_app"
        with pytest.raises(
            DFASpecificationError,
            match="ABCI app class name should end in `AbciApp`; ABCI app name found `abci_app`",
        ):
            DFA(**dfa_kwargs)  # type: ignore

        with pytest.raises(
            DFASpecificationError,
            match="Round class name should end in `Round`; Round app name found `StateBRoun`",
        ):
            DFA(
                label="DummyAbciApp",
                states={"StateARound", "StateBRoun"},
                default_start_state="StateARound",
                start_states={"StateARound"},
                final_states={"StateBRoun"},
                alphabet_in={"event_b"},
                transition_func={
                    ("StateARound", "event_b"): "StateBRoun",
                },
            )



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2021-2023 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""This module contains utility functions for the 'abstract_round_abci' skill."""

import builtins
import collections
import dataclasses
import sys
import types
import typing
from hashlib import sha256
from math import ceil
from typing import (
    Any,
    Dict,
    FrozenSet,
    Iterator,
    List,
    Optional,
    Set,
    Tuple,
    Type,
    TypeVar,
    Union,
    cast,
)
from unittest.mock import MagicMock

import typing_extensions
from eth_typing.bls import BLSPubkey, BLSSignature
from py_ecc.bls import G2Basic as bls
from typing_extensions import Literal, TypeGuard, TypedDict


MAX_UINT64 = 2**64 - 1
DEFAULT_TENDERMINT_P2P_PORT = 26656


class VerifyDrand:  # pylint: disable=too-few-public-methods
    """
    Tool to verify Randomness retrieved from various external APIs.

    The ciphersuite used is BLS_SIG_BLS12381G2_XMD:SHA-256_SSWU_RO_NUL_

    cryptographic-specification section in https://drand.love/docs/specification/
    https://github.com/ethereum/py_ecc
    """

    @classmethod
    def _int_to_bytes_big(cls, value: int) -> bytes:
        """Convert int to bytes."""
        if value < 0 or value > MAX_UINT64:
            raise ValueError(
                "VerifyDrand can only handle positive numbers representable with 8 bytes"
            )
        return int.to_bytes(value, 8, byteorder="big", signed=False)

    @classmethod
    def _verify_randomness_hash(cls, randomness: bytes, signature: bytes) -> bool:
        """Verify randomness hash."""
        return sha256(signature).digest() == randomness

    @classmethod
    def _verify_signature(
        cls,
        pubkey: Union[BLSPubkey, bytes],
        message: bytes,
        signature: Union[BLSSignature, bytes],
    ) -> bool:
        """Verify randomness signature."""
        return bls.Verify(
            cast(BLSPubkey, pubkey), message, cast(BLSSignature, signature)
        )

    def verify(self, data: Dict, pubkey: str) -> Tuple[bool, Optional[str]]:
        """
        Verify drand value retried from external APIs.

        :param data: dictionary containing drand parameters.
        :param pubkey: league of entropy public key
                       public-endpoints section in https://drand.love/developer/http-api/
        :returns: bool, error message
        """

        encoded_pubkey = bytes.fromhex(pubkey)
        try:
            randomness = data["randomness"]
            signature = data["signature"]
            round_value = int(data["round"])
        except KeyError as e:
            return False, f"DRAND dict is missing value for {e}"

        previous_signature = data.pop("previous_signature", "")
        encoded_randomness = bytes.fromhex(randomness)
        encoded_signature = bytes.fromhex(signature)
        int_encoded_round = self._int_to_bytes_big(round_value)
        encoded_previous_signature = bytes.fromhex(previous_signature)

        if not self._verify_randomness_hash(encoded_randomness, encoded_signature):
            return False, "Failed randomness hash check."

        msg_b = encoded_previous_signature + int_encoded_round
        msg_hash_b = sha256(msg_b).digest()

        if not self._verify_signature(encoded_pubkey, msg_hash_b, encoded_signature):
            return False, "Failed bls.Verify check."

        return True, None


def get_data_from_nested_dict(
    nested_dict: Dict, keys: str, separator: str = ":"
) -> Any:
    """Gets content from a nested dictionary, using serialized response keys which are split by a given separator.

    :param nested_dict: the nested dictionary to get the content from
    :param keys: the keys to use on the nested dictionary in order to get the content
    :param separator: the separator to use in order to get the keys list.
    Choose the separator carefully, so that it does not conflict with any character of the keys.

    :returns: the content result
    """
    parsed_keys = keys.split(separator)
    for key in parsed_keys:
        nested_dict = nested_dict[key]
    return nested_dict


def get_value_with_type(value: Any, type_name: str) -> Any:
    """Get the given value as the specified type."""
    return getattr(builtins, type_name)(value)


def parse_tendermint_p2p_url(url: str) -> Tuple[str, int]:
    """Parse tendermint P2P url."""
    hostname, *_port = url.split(":")
    if len(_port) > 0:
        port_str, *_ = _port
        port = int(port_str)
    else:
        port = DEFAULT_TENDERMINT_P2P_PORT

    return hostname, port


##
# Typing utils - to be extracted to open-aea
##


try:
    # Python >=3.8 should have these functions already
    from typing import get_args as _get_args  # pylint: disable=ungrouped-imports
    from typing import get_origin as _get_origin  # pylint: disable=ungrouped-imports
except ImportError:  # pragma: nocover
    # Python 3.7
    def _get_origin(tp):  # type: ignore
        """Copied from the Python 3.8 typing module"""
        if isinstance(tp, typing._GenericAlias):  # pylint: disable=protected-access
            return tp.__origin__
        if tp is typing.Generic:
            return typing.Generic
        return None

    def _get_args(tp):  # type: ignore
        """Copied from the Python 3.8 typing module"""
        if isinstance(tp, typing._GenericAlias):  # pylint: disable=protected-access
            res = tp.__args__
            if get_origin(tp) is collections.abc.Callable and res[0] is not Ellipsis:
                res = (list(res[:-1]), res[-1])
            return res
        return ()


def get_origin(tp):  # type: ignore
    """
    Get the unsubscripted version of a type.

    This supports generic types, Callable, Tuple, Union, Literal, Final and
    ClassVar. Returns None for unsupported types.
    Examples:
        get_origin(Literal[42]) is Literal
        get_origin(int) is None
        get_origin(ClassVar[int]) is ClassVar
        get_origin(Generic) is Generic
        get_origin(Generic[T]) is Generic
        get_origin(Union[T, int]) is Union
        get_origin(List[Tuple[T, T]][int]) == list
    """
    return _get_origin(tp)


def get_args(tp):  # type: ignore
    """
    Get type arguments with all substitutions performed.

    For unions, basic simplifications used by Union constructor are performed.
    Examples:
        get_args(Dict[str, int]) == (str, int)
        get_args(int) == ()
        get_args(Union[int, Union[T, int], str][int]) == (int, str)
        get_args(Union[int, Tuple[T, int]][str]) == (int, Tuple[str, int])
        get_args(Callable[[], T][int]) == ([], int)
    """
    return _get_args(tp)


##
# The following is borrowed from https://github.com/tamuhey/dataclass_utils/blob/81580d2c0c285081db06be02b4ecdd125532bef5/dataclass_utils/type_checker.py#L152
##


def is_pep604_union(ty: Type[Any]) -> bool:
    """Check if a type is a PEP 604 union."""
    return sys.version_info >= (3, 10) and ty is types.UnionType  # type: ignore # noqa: E721


def _path_to_str(path: List[str]) -> str:
    """Convert a path to a string."""
    return " -> ".join(reversed(path))


class AutonomyTypeError(TypeError):
    """Type Error for the Autonomy type check system."""

    def __init__(
        self,
        ty: Type[Any],
        value: Any,
        path: Optional[List[str]] = None,
    ):
        """Initialize AutonomyTypeError."""
        self.ty = ty
        self.value = value
        self.path = path or []
        super().__init__()

    def __str__(self) -> str:
        """Get string representation of AutonomyTypeError."""
        path = _path_to_str(self.path)
        msg = f"Error in field '{path}'. Expected type {self.ty}, got {type(self.value)} (value: {self.value})"
        return msg


Result = Optional[AutonomyTypeError]  # returns error context


def check(  # pylint: disable=too-many-return-statements
    value: Any, ty: Type[Any]
) -> Result:
    """
    Check a value against a type.

    # Examples
    >>> assert is_error(check(1, str))
    >>> assert not is_error(check(1, int))
    >>> assert is_error(check(1, list))
    >>> assert is_error(check(1.3, int))
    >>> assert is_error(check(1.3, Union[str, int]))
    """
    if isinstance(value, MagicMock):
        # testing - any magic value is ignored
        return None
    if not isinstance(value, type) and dataclasses.is_dataclass(ty):
        # dataclass
        return check_dataclass(value, ty)
    if is_typeddict(ty):
        # should use `typing.is_typeddict` in future
        return check_typeddict(value, ty)
    to = get_origin(ty)
    if to is not None:
        # generics
        err = check(value, to)
        if is_error(err):
            return err

        if to is list or to is set or to is frozenset:
            err = check_mono_container(value, ty)
        elif to is dict:
            err = check_dict(value, ty)  # type: ignore
        elif to is tuple:
            err = check_tuple(value, ty)
        elif to is Literal:
            err = check_literal(value, ty)
        elif to is Union or is_pep604_union(to):
            err = check_union(value, ty)
        elif to is type:
            err = check_class(value, ty)
        return err
    if isinstance(ty, type):
        # concrete type
        if is_pep604_union(ty):
            pass  # pragma: no cover
        elif issubclass(ty, bool):
            if not isinstance(value, ty):
                return AutonomyTypeError(ty=ty, value=value)
        elif issubclass(ty, int):  # For boolean
            return check_int(value, ty)
        elif not isinstance(value, ty):
            return AutonomyTypeError(ty=ty, value=value)
    return None


def check_class(value: Any, ty: Type[Any]) -> Result:
    """Check class type."""
    if not issubclass(value, get_args(ty)):
        return AutonomyTypeError(ty=ty, value=value)
    return None


def check_int(value: Any, ty: Type[Any]) -> Result:
    """Check int type."""
    if isinstance(value, bool) or not isinstance(value, ty):
        return AutonomyTypeError(ty=ty, value=value)
    return None


def check_literal(value: Any, ty: Type[Any]) -> Result:
    """Check literal type."""
    if all(value != t for t in get_args(ty)):
        return AutonomyTypeError(ty=ty, value=value)
    return None


def check_tuple(value: Any, ty: Type[Tuple[Any, ...]]) -> Result:
    """Check tuple type."""
    types_ = get_args(ty)
    if len(types_) == 2 and types_[1] == ...:
        # arbitrary length tuple (e.g. Tuple[int, ...])
        for v in value:
            err = check(v, types_[0])
            if is_error(err):
                return err
        return None

    if len(value) != len(types_):
        return AutonomyTypeError(ty=ty, value=value)
    for v, t in zip(value, types_):
        err = check(v, t)
        if is_error(err):
            return err
    return None


def check_union(value: Any, ty: Type[Any]) -> Result:
    """Check union type."""
    if any(not is_error(check(value, t)) for t in get_args(ty)):
        return None
    return AutonomyTypeError(ty=ty, value=value)


def check_mono_container(
    value: Any, ty: Union[Type[List[Any]], Type[Set[Any]], Type[FrozenSet[Any]]]
) -> Result:
    """Check mono container type."""
    ty_item = get_args(ty)[0]
    for v in value:
        err = check(v, ty_item)
        if is_error(err):
            return err
    return None


def check_dict(value: Dict[Any, Any], ty: Type[Dict[Any, Any]]) -> Result:
    """Check dict type."""
    args = get_args(ty)
    ty_key = args[0]
    ty_item = args[1]
    for k, v in value.items():
        err = check(k, ty_key)
        if is_error(err):
            return err
        err = check(v, ty_item)
        if err is not None:
            err.path.append(k)
            return err
    return None


def check_dataclass(value: Any, ty: Type[Any]) -> Result:
    """Check dataclass type."""
    if not dataclasses.is_dataclass(value):
        return AutonomyTypeError(ty, value)
    for k, ty_ in typing.get_type_hints(ty).items():
        v = getattr(value, k)
        err = check(v, ty_)
        if err is not None:
            err.path.append(k)
            return err
    return None


def check_typeddict(value: Any, ty: Type[Any]) -> Result:
    """Check typeddict type."""
    if not isinstance(value, dict):
        return AutonomyTypeError(ty, value)  # pragma: no cover
    is_total: bool = ty.__total__  # type: ignore
    for k, ty_ in typing.get_type_hints(ty).items():
        if k not in value:
            if is_total:
                return AutonomyTypeError(ty_, value, [k])
            continue
        v = value[k]
        err = check(v, ty_)
        if err is not None:
            err.path.append(k)
            return err
    return None


# TODO: incorporate
def is_typevar(ty: Type[Any]) -> TypeGuard[TypeVar]:
    """Check typevar."""
    return isinstance(ty, TypeVar)  # pragma: no cover


def is_error(ret: Result) -> TypeGuard[AutonomyTypeError]:
    """Check error."""
    return ret is not None


def is_typeddict(ty: Type[Any]) -> TypeGuard[Type[TypedDict]]:  # type: ignore
    """Check typeddict."""
    # TODO: Should use `typing.is_typeddict` in future
    #       or, use publich API
    T = "_TypedDictMeta"
    for mod in [typing, typing_extensions]:
        if hasattr(mod, T) and isinstance(ty, getattr(mod, T)):
            return True
    return False


def check_type(name: str, value: Any, type_hint: Any) -> None:
    """Check value against type hint recursively"""
    err = check(value, type_hint)
    if err is not None:
        err.path.append(name)
        raise err


def is_primitive_or_none(obj: Any) -> bool:
    """Checks if the given object is a primitive type or `None`."""
    primitives = (bool, int, float, str)
    return isinstance(obj, primitives) or obj is None


def is_json_serializable(obj: Any) -> bool:
    """Checks if the given object is json serializable."""
    if isinstance(obj, (tuple, list)):
        return all(is_json_serializable(x) for x in obj)
    if isinstance(obj, dict):
        return all(
            is_primitive_or_none(k) and is_json_serializable(v) for k, v in obj.items()
        )

    return is_primitive_or_none(obj)


def filter_negative(mapping: Dict[str, int]) -> Iterator[str]:
    """Return the keys of a dictionary for which the values are negative integers."""
    return (key for key, number in mapping.items() if number < 0)


def consensus_threshold(nb: int) -> int:
    """
    Get consensus threshold.

    :param nb: the number of participants
    :return: the consensus threshold
    """
    return ceil((2 * nb + 1) / 3)
